# Ajuste no lineal de elipses
# del artículo
# Least-squares orthogonal distances fitting of circle,
#     sphere, ellipse, hyperbola, and parabola
# Sung Joon Ahn, Wolfgang Rauh, Hans-Jurgen Warnecke,
# Pattern Recognition 34 (2001) 2283-2303

import sys
import numpy as np
import portogonal
import retro

# La elipse inicial la pasamos en la línea de comandos
n = len( sys.argv )
if n != 6 :
	print( "Args: Xc Yc a b alpha" )
	sys.exit(1)

Xc = float( sys.argv[1] )
Yc = float( sys.argv[2] )
a = float( sys.argv[3] )
b = float( sys.argv[4] )
alpha = float( sys.argv[5] )

va = np.array( [ [Xc], [Yc], [a], [b], [alpha] ] )

# Leer los puntos de entrada de un archivo
Pi = np.loadtxt( "puntos.txt" )
print( Pi.shape )
# Pi es una matrix de tamaño  n x 2
n, m = Pi.shape

# Los puntos de entrada se transforman
# hacia la elipse en el origen
# pi es una matrix de tamaño  n x 2
pi = np.zeros( (n, 2 ) )  # Los puntos de entrada en el sistema de coordenadas
                          # centrado en la elipse en el origen.

ppi = np.zeros( (n, 2 ) ) # Los puntos sobre la elipse

R = np.zeros( (2, 2 ) )
Rinv = np.zeros( (2, 2 ) )

Q = np.zeros( (2, 2 ) ) 
Qinv = np.zeros( (2, 2 ) ) 

B = np.zeros( (2, 5 ) ) 
J = np.zeros( (2*n, 6 ) ) 
Cx = np.zeros( (2,1) )
Xi = np.zeros( (2,1) )
vv = np.zeros( (2,1) )
vu = np.zeros( (2,1) )
vw = np.zeros( (2,1) )

# El Newton para resolver el ajuste de la elipse
i = 0
while i < 30 :
	Xc = Cx[0,0] = va[0,0]
	Yc = Cx[1,0] = va[1,0]
	a = va[2,0]
	b = va[3,0]
	c = np.cos( va[4,0] )
	s = np.sin( va[4,0] )
	R[0,0] = c
	R[1,1] = R[0,0]
	R[0,1] = s
	R[1,0] = -R[0,1]
	Rinv = R.T

	j = 0
	while j < n :
		vw[0,0] = Pi[j,0]
		vw[1,0] = Pi[j,1]
		vw -= Cx
		vv = R @ vw 
		pi[j,:] = vv.T
		j += 1

	# Calculamos los puntos sobre la elipse
	j = 0
	while j < n :
		ppi[j:] = portogonal.punto_ortogonal( a, b, pi[j,0], pi[j,1] )
		j += 1

	# Contruimos la Jacobiana
	# es de tamaño 2*n x 5
	j = 0
	while j < n :
		xi = pi[ j,0 ]
		yi = pi[ j,1 ]

		x = ppi[ j,0 ]
		y = ppi[ j,1 ]

		a2 = a*a
		b2 = b*b

		Q[0,0] = b2*x
		Q[0,1] = a2*y

		Q[1,0] = (a2-b2)*y + b2*yi
		Q[1,1] = (a2-b2)*x - a2*xi 

		Qinv = np.linalg.inv( Q )

		# Matrix B

		B[0,0] = b2*x*c - a2*y*s
		B[1,0] = b2*(yi-y)*c + a2*(xi - x)*s

		B[0,1] = b2*x*s + a2*y*c
		B[1,1] = b2*(yi-y)*s - a2*(xi - x)*c

		B[0,2] = a*(b2 - y*y)
		B[1,2] = 2.0*a*y*(xi-x)

		B[0,3] = b*(a2 - x*x)
		B[1,3] = -2.0*b*x*(yi - y)

		B[0,4] = (a2 - b2)*x*y
		B[1,4] = (a2 - b2)*(x*x  - y*y - x*xi + y*yi) 

		J[2*j:2*j+2,0:5] = Rinv @ Qinv @ B

		j += 1

	j = 0
	while j < n :
		vv[0,0] = ppi[j,0]
		vv[1,0] = ppi[j,1]

		Xi[:,0] = (Rinv @ vv + Cx)[:,0]

		vu[0,0] = Pi[j,0]
		vu[1,0] = Pi[j,1]

		vv[:,0] = vu[:,0] - Xi[:,0]

		J[ 2*j, 5] = vv[0,0]
		J[ 2*j+1, 5] = vv[1,0]
		
		j += 1
	#   2x2        2x1

	# Calculamos Deltaa
	# 16 x 5
	# La jacobiana invertida 5 x 16 
	# Xpp tiene tamaño 16 x 1
	#                 5 x 16      16 x 1
	Rtriang = np.linalg.qr( J, mode='r' )

	# Aplicar la retrosustitución
	Deltaa = retro.retrosustitucion( Rtriang, Rtriang[:,5] )
	Deltaa = Deltaa.reshape( (-1,1) )
	
	# ¿Cómo paramos el algoritmo?
	v = np.linalg.norm( Deltaa )
	print( "Paro:", v )
	if v < 1e-10 :
		break

	va += Deltaa
	print( i, va[0,0], va[1,0], va[2,0], va[3,0], va[4,0] )

	i += 1
