#
# L.G. de la Fraga, I. Vite Silva, and N. Cruz-Cortés,
# Euclidean distance fit of conics using differential evolution,
# Book Chapter in: Evolutionary Image Analysis and Signal Processing
# Studies in Computational Intelligence 2013,
# S. Gagnoni (Ed.)
# 2009, Springer 
#
import sys
import math

def pprime( va, p ) :
	a = va[2]
	b = va[3]
	x = p[0]
	y = p[1]

	# Inicialización:
	ai_1 = math.sqrt( x*x + y*y )
	if ai_1 < 1e-6 :
		ai_1 = math.pi/2.0

	v = a*a - b*b

	if math.fabs(y) < b and math.fabs(x) <= v/a :
		if math.fabs(y) < 1e-6 :
			ai_1 = math.pi/2.0
		else :
			p = math.fabs(x)/(1.0 + math.fabs(y)/ai_1)
			if x < 0 :
				p = -p
			ai_1 = math.atan2( a*y, b*(x-p) )
	else :
		ai_1 = math.atan2( a*y, b*x )

	t = ai_1

	i=1
	while i < 10 :
		c = math.cos(t)
		s = math.sin(t)
		v = a*a - b*b
		f = y*b*c - x*a*s + (a*a - b*b)*s*c
		fp  = -y*b*s - x*a*c - v*s*s + v*c*c
		Delta_t = -f/fp

		if abs(Delta_t) < 1e-8 :
			break

		t += Delta_t

		i += 1

	return ( a*math.cos( t ),  b*math.sin( t ) )

def backSubstitution( R, n, v ) :
	# /* Backsustitution */
	# R tiene dimensiones n x n
	# 0 1 2 3 4 5
	i = n-2  # 4

	if abs(R[i,i]) < 1e-10:
		v[i] = 0.0
	else :
		v[i] = R[i,i+1]/R[i,i]

	m = n-1
	i = n-3
	while i>=0 :
		suma = 0.0
		j = i+1
		while j<m :
			suma += R[i][j] * v[j];
			j += 1

		v[i] = ( R[i,m] - suma )/R[i,i];
		i -= 1
